""" ssm.api.environment

Core abstraction for environment-aware boto, which
reduces a lot of boilerplate with boto sessions/profiles
"""

import collections

import boto3

from ssm import abcs, util

from .manager import SecretManager

# default logger only used for class-methods,
# because for this there is no instance is available
LOGGER = util.get_logger(__name__)


class Environment(
    abcs.Loggable,
):
    """ """

    @staticmethod
    def normalize_env_name(name):
        """ """
        return name

    @classmethod
    def from_profile(cls, name):
        """
        instantiates an Environment from AWS profile name,
        preferably something like 'dev', or 'datacenter1',
        or 'prod'.
        """
        assert isinstance(name, str), type(name)
        normal_name = Environment.normalize_env_name(name)
        try:
            config = cls.ENV_CONFIGS[normal_name]
        except KeyError:
            err = "no such environment name `{}` found inside config:\n  {}"
            raise KeyError(err.format(normal_name, cls.ENV_CONFIGS))
        return cls(config=config)

    @property
    def profile_name(self):
        """
        returns the default profile name from config
        """
        return self.config.get("profile_name")

    profile = profile_name

    @property
    def account_id(self):
        return self.config.get("account_id")

    account = account_id

    @property
    def region_name(self):
        """
        returns the default region name from config
        """
        return self.config.get("region_name") or getattr(
            getattr(self, "session", None), "region_name", None
        )

    region = region_name

    def __init__(self, config={}, **kwargs):
        if not isinstance(config, dict):
            err = "expected dict for `config`, got {}"
            raise ValueError(err.format(config))
        self.config = config
        self.name = self.config["name"]
        self.logger_name = f"A[{self.name}]"
        self.init_session()
        super().__init__(**kwargs)

    def get_boto(self, client_type):
        """
        get a classic boto client.  this is still
        necessary for example with some route53 operations
        """
        import boto

        fxn = getattr(boto, f"connect_{client_type}")
        creds = self.session.get_credentials()
        return fxn(
            aws_access_key_id=creds.access_key,
            aws_secret_access_key=creds.secret_key,
            security_token=creds.token,
        )

    def init_session(self):
        """
        initializes boto session and various service handles
        """
        self.session = boto3.session.Session(
            profile_name=self.profile_name, region_name=self.region_name
        )
        self.s3 = self.session.client("s3")
        self.ec2 = self.session.client("ec2")
        self.emr = self.session.client("emr")
        self.ssm = self.session.client("ssm")
        self.iam = self.session.client("iam")
        self.sts = self.session.client("sts")
        self.cloudwatch = self.session.client("cloudwatch")
        self.route53 = self.session.client("route53")
        # always us-east-1
        self.support = boto3.session.Session(
            profile_name=self.profile_name, region_name="us-east-1"
        ).client("support")
        self.redshift = self.session.client("redshift")
        self.cloudformation = self.session.client("cloudformation")

    @property
    def user_names(self):
        """ """
        self.logger.debug("computing IAM usernames for this environment")
        return [u["UserName"] for u in self.iam.list_users()["Users"]]

    @property
    def role_names(self):
        """ """
        self.logger.debug("computing IAM roles for this environment")
        return [u["RoleName"] for u in self.iam.list_roles()["Roles"]]

    def has_role(self, name):
        """ """
        return name in self.role_names

    def has_user(self, name):
        """ """
        return name in self.user_names

    def __eq__(self, other):
        """environments are equal if their names are equal"""
        return all([isinstance(other, (self.__class__,)), self.name == other.name])

    def __hash__(self):
        return hash(self.name)

    def __ne__(self, other):
        """opposite of self==other, otherwise this would use obj id's"""
        return not self == other

    def __str__(self):
        """
        human friendly strings for things like console
        output, i.e. with `secrets shell`, or logging
        """
        return f"<Environment: {self.name} @ {self.profile_name}>"

    __repr__ = __str__

    @property
    def secrets(self):
        """
        returns a secrets manager for this session
        (see ssm for more info)
        """
        return SecretManager(env=self)


from botocore import session

Environment.ENV_CONFIGS = collections.OrderedDict()
Environment.ALL_PROFS = session.Session().available_profiles
tmp = "loading environments from {} available profiles.."
LOGGER.debug(tmp.format(len(Environment.ALL_PROFS)))
for profile_name in Environment.ALL_PROFS:
    # LOGGER.debug("\t{}".format(profile_name))
    Environment.ENV_CONFIGS[profile_name] = dict(
        profile_name=profile_name,
    )
LOGGER.debug("normalizing and pre-computing env metadata")
for env_name, env_config in Environment.ENV_CONFIGS.copy().items():
    LOGGER.debug(f"\t- {env_name}")
    env_config["name"] = Environment.normalize_env_name(env_name)
LOGGER.debug("done loading environment configs")



""" ssm.cli.options (boilerplate for click)

    Common CLI arguments for reuse
"""

from functools import partial

import click

cluster = click.option(
    "--cluster",
    required=False,
    default="",
    help="cluster to use (defaults to autodetection from env)",
)
comment = click.option("--comment", help="comment or description", required=False)
container = click.option("--container", help="container to enter", required=False)
key = click.option("--key", help="key to use", default="", required=False)

top = click.option(
    "--top", help="triage and only describe top N results", required=False
)
existing_file = click.option("--file", type=click.Path(exists=True))
filter_service_partial = partial(
    click.option, "--service", help="filter for service", required=False
)
filter_service = filter_service_partial(default=".*")

file_format_partial = partial(
    click.option,
    "--format",
    type=click.Choice(["json", "yaml", "yml", "env"]),
    help="file format",
)
file_format = format = file_format_partial(required=True)
file_format_yml_default = file_format_yaml_default = file_format_partial(
    required=False, default="yaml"
)
required_dest_bucket = dest_bucket = click.option(
    "--dest-bucket", envvar="DEST_BUCKET", help="dest bucket name", required=True
)


optional_user = user = click.option(
    "--user",
    help="username (default will attempt auto-detect)",
    required=False,
    default="",
)
optional_users = users = click.option(
    "--users", help="user list (comma-separatted)", required=False, default=""
)
optional_database = database = click.option(
    "--database",
    help="database (default will attempt auto-detect)",
    required=False,
    default="",
)

required_bucket = click.option("--bucket", help="bucket name", required=True)

raw = click.option(
    "--raw", "-r", help="unquotes text return values (like jq -r)", default=False
)

force_or_dry = force = dry_run = click.option(
    "--force/--dry-run", "-f/-d", help="perform actions", default=False, envvar="FORCE"
)

pause_or_nowait = no_wait = pause = click.option(
    "--pause/--no-wait", help="pause before starting", default=True
)

optional_prefix = click.option(
    "--prefix", help="src prefix to operate under", required=False, default=""
)

dest_prefix = click.option(
    "--dest-prefix",
    envvar="DEST_PREFIX",
    help="dest prefix to operate under",
    required=False,
    default="",
)

required_user = click.option(
    "--user", help="username to use", required=True, default=""
)

optional_src_bucket = click.option(
    "--src-bucket", envvar="SRC_BUCKET", help="src bucket name", required=False
)

required_src_bucket = click.option(
    "--src-bucket", envvar="SRC_BUCKET", help="src bucket name", required=True
)

command = click.option(
    "--command",
    "-c",
    required=False,
    default="",
    help="Command to run (like bash -c)",
)
required_command = click.option(
    "--command",
    "-c",
    default="",
    help="Command to run (like bash -c)",
)

script = click.option("--script", "-s", help="Script to run", default="")

cascade_partial = partial(
    click.option,
    "--cascade/--no-cascade",
    "-c/",
    required=False,
    help="cascading search",
)
cascade = cascade_partial(default=True)
no_cascade = cascade_partial(default=False)

no_cascade = click.option(
    "--no-cascade",
    required=False,
    default=False,
    flag=True,
    help="regex to filter results by",
)

filter = click.option(
    "--filter",
    required=False,
    default=".*",
    help="regex to filter results by",
)

optional_arn = click.option(
    "--arn",
    required=False,
    default="",
    help="ARN",
)
optional_role_name = click.option(
    "--role-name",
    required=False,
    default="",
    help="Role name",
)

env_partial = partial(
    click.option,
    "--env",
    envvar="AWS_PROFILE",
    help="Environment to use",
)
src_env_partial = partial(
    click.option,
    "--src-env",
    help="Source Environment name",
)
dest_env_partial = partial(
    click.option,
    "--dest-env",
    help="Destination Environment name",
)

required_env = env_partial(
    required=True,
)
optional_env = env_no_default = env_partial(
    required=False,
)
env = default_env = env_partial(
    default="default",
    required=False,
)
src_env = src_env_default = src_env_partial(default="default", required=False)
dest_env = dest_env_default = dest_env_partial(default="default", required=False)
src_env_no_default = src_env_partial(default=None, required=False)
dest_env_no_default = dest_env_partial(default=None, required=False)

required_stack = click.option(
    "--stack",
    required=True,
    help="Stack name to use",
)

required_key = click.option(
    "--key",
    required=True,
    help="Key name to use",
)

src_prefix = click.option(
    "--src-prefix",
    envvar="SRC_PREFIX",
    help="src prefix to operate under",
    required=False,
    default="",
)

required_regexp = click.option(
    "--regexp",
    required=True,
    help="RegExp to use",
)



""" ssm.util
"""
import logging
import functools

import termcolor
import coloredlogs

blue = functools.partial(termcolor.colored, color="blue")
red = functools.partial(termcolor.colored, color="red")
green = functools.partial(termcolor.colored, color="green")
yellow = functools.partial(termcolor.colored, color="yellow")
bold = functools.partial(termcolor.colored, attrs=["bold"])


def is_string(obj):
    return isinstance(obj, str)


def fatal_error(msg):
    """ """
    LOGGER.info("{} {}".format(red("error:"), msg))
    raise SystemExit(1)


def get_logger(name):
    """
    utility function for returning a logger
    with standard formatting patterns, etc
    """

    class DuplicateFilter(logging.Filter):
        def filter(self, record):
            # add other fields if you need more granular comparison, depends on your app
            current_log = (record.module, record.levelno, record.msg)
            if current_log != getattr(self, "last_log", None):
                self.last_log = current_log
                return True
            return False

    formatter = coloredlogs.ColoredFormatter(
        fmt=" - ".join(
            [
                # "[%(asctime)s]",
                "%(levelname)s\t",
                "%(name)s\t",
                "%(message)s",
            ]
        ),
        # datefmt="%Y-%m-%d %H:%M:%S",
    )
    log_handler = logging.StreamHandler()
    log_handler.setFormatter(formatter)
    logger = logging.getLogger(name)
    if not logger.handlers:
        # prevents duplicate registration
        logger.addHandler(log_handler)
    logger.addFilter(DuplicateFilter())  # add the filter to it
    # FIXME: get this from some kind of global config
    logger.setLevel("DEBUG")
    return logger


LOGGER = get_logger(__name__)



""" ssm.cli.wrapper (click boilerplate)
    Reusable wrapper for building CLIs
"""

import functools

import click

from ssm import abcs, util

LOGGER = util.get_logger(__file__)


class ApiWrapper(abcs.Loggable):
    """
    a wrapper that turns a API function into a click CLI subcommand
    """

    BASE_OPTIONS = [
        # every command gets the --debug option
        click.option(
            "--debug",
            default=False,
            is_flag=True,
            help="Enables verbose mode.",
        ),
    ]

    def __init__(
        self,
        command_name=None,
        publishers=[],
        subcommand_name=None,
        fxn=None,
        extra_options=None,
        aliases=[],
        help=None,
        entry=None,
    ):
        self.entry = entry
        self.aliases = aliases
        self.is_subcommand = isinstance(self.entry, (click.core.Group,))
        self.is_stand_alone = self.entry is None
        self.subcommand_name = self.name = subcommand_name or fxn.__name__.replace(
            "_", "-"
        )
        self.command_name = command_name
        self.fxn = fxn
        self.fxn.publishers = getattr(self.fxn, "publishers", publishers)
        self.extra_options = extra_options
        default_help = "no docstring"
        self.help = help or getattr(fxn, "__doc__", default_help) or default_help
        self.help = self.help.strip()
        self.proxy = self.get_proxy()
        if not callable(self.proxy):
            err = f"Expected callable for proxy, got '{self.proxy}'"
            raise ValueError(err)
        if not (self.is_subcommand or self.is_stand_alone):
            err = (
                "expected a group or a standalone "
                "command, got {} of type {} for entry"
            )
            err = err.format(self.entry, type(self.entry))
            raise ValueError(err)
        super().__init__()

    def get_proxy(self):
        """ """
        options = self.extra_options
        if self.is_subcommand:
            # otherwise base options would be added twice for stand-alone style CLIs
            options += self.__class__.BASE_OPTIONS

        @functools.wraps(self.fxn)
        def proxy(*args, **kwargs):
            """ """
            # if args and isinstance(args[0],(click.core.Context,)):
            #     # none of the api commands are anticipating the click context
            #     args = args[1:]
            args = [x for x in args if not isinstance(x, (click.core.Context,))]
            LOGGER.debug(f"proxying args={args}, kwargs={kwargs}")
            api_result = self.fxn(*args, **kwargs)
            print(api_result)
            if not isinstance(api_result, (dict, list)):
                LOGGER.warning("api result returned was not JSON!")
            # if util.is_string(api_result):
            #     print(result)
            # else:
            #     print(json.dumps(api_result, indent=2))
            return api_result

        for option in options:
            proxy = option(proxy)

        if self.is_subcommand:
            if self.aliases:
                result = self.entry.command(
                    name=self.name, help=self.help, aliases=self.aliases
                )(proxy)
            else:
                result = self.entry.command(name=self.name, help=self.help)(proxy)
        elif self.entry is None:
            result = click.command()(proxy)
        else:
            err = "unknown entry type '{}' for '{}'".format(
                type(self.entry), self.entry
            )
            LOGGER.critical(err)
            raise RuntimeError(err)
        # print(result)
        return result
